#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/kglab#license-and-copyright

"""
SPARQL query abstractions.
"""

import re
import typing

import pandas as pd  # type: ignore  # pylint: disable=E0401
import pyvis  # type: ignore  # pylint: disable=E0401
import rdflib  # type: ignore  # pylint: disable=E0401

from ..pkg_types import RDF_Node
from ..gpviz import GPViz
from ..util import get_gpu_count

from .base import Queryable


## pre-constructor set-up
if get_gpu_count() > 0:
    import cudf  # type: ignore  # pylint: disable=E0401


class SparqlQueryable (Queryable):
    """
SPARQL implementation for Queryable abstract class.
    """

    def __init__ (
        self,
        kg: typing.Any,
        ) -> None:
        """
Constructor.
        """
        self.kg = kg


    ######################################################################
    ## SPARQL queries

    def query (  # pylint: disable=W0221
        self,
        query: str,
        *,
        bindings: typing.Optional[ dict ] = None,
        ) -> typing.Iterable:
        """
Wrapper for [`rdflib.Graph.query()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=query#rdflib.Graph.query) to perform a SPARQL query on the RDF graph.

    query:
text for the SPARQL query

    bindings:
initial variable bindings

    yields:
[`rdflib.query.ResultRow`](https://rdflib.readthedocs.io/en/stable/_modules/rdflib/query.html?highlight=ResultRow#) named tuples, to iterate through the query result set
        """
        if not bindings:
            bindings = {}

        yield from self.kg._g.query(  # pylint: disable=W0212
                query,
                initBindings = bindings,
            )


    def query_as_df (  # pylint: disable=W0221
        self,
        query: str,
        *,
        bindings: typing.Optional[ dict ] = None,
        simplify: bool = True,
        pythonify: bool = True,
        ) -> pd.DataFrame:
        """
Wrapper for [`rdflib.Graph.query()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=query#rdflib.Graph.query) to perform a SPARQL query on the RDF graph.

    query:
text for the SPARQL query

    bindings:
initial variable bindings

    simplify:
convert terms in each row of the result set into a readable representation for each term, using N3 format

    pythonify:
convert instances of [`rdflib.term.Literal`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Literal#rdflib.term.Identifier) to their Python literal representation

    returns:
the query result set represented as a [`pandas.DataFrame`](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html); uses the [RAPIDS `cuDF` library](https://docs.rapids.ai/api/cudf/stable/) if GPUs are enabled
        """
        if not bindings:
            bindings = {}

        row_iter = self.kg._g.query(query, initBindings=bindings)  # pylint: disable=W0212

        if simplify:
            rows_list = [ self.n3fy_row(r.asdict(), pythonify=pythonify) for r in row_iter ]
        else:
            rows_list = [ r.asdict() for r in row_iter ]

        if self.kg.use_gpus:
            df = cudf.DataFrame(rows_list)  # pylint: disable=E0606
        else:
            df = pd.DataFrame(rows_list)

        return df


    def n3fy (
        self,
        node: RDF_Node,
        *,
        pythonify: bool = True,
        ) -> typing.Any:
        """
Wrapper for RDFlib [`n3()`](https://rdflib.readthedocs.io/en/stable/utilities.html?highlight=n3#serializing-a-single-term-to-n3) and [`toPython()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=toPython#rdflib.Variable.toPython) to serialize a node into a human-readable representation using N3 format.

    node:
must be a [`rdflib.term.Node`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Node#rdflib.term.Node)

    pythonify:
flag to force instances of [`rdflib.term.Literal`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Literal#rdflib.term.Identifier) to their Python literal representation

    returns:
text (or Python objects) for the serialized node
        """
        if pythonify and isinstance(node, rdflib.term.Literal):
            serialized = node.toPython()
        else:
            serialized = node.n3(self.kg._g.namespace_manager)  # type: ignore  # pylint: disable=W0212

        return serialized


    def n3fy_row (
        self,
        row_dict: dict,
        *,
        pythonify: bool = True,
        ) -> dict:
        """
Wrapper for RDFlib [`n3()`](https://rdflib.readthedocs.io/en/stable/utilities.html?highlight=n3#serializing-a-single-term-to-n3) and [`toPython()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=toPython#rdflib.Variable.toPython) to serialize one row of a result set from a SPARQL query into a human-readable representation for each term using N3 format.

    row_dict:
one row of a SPARQL query results, as a `dict`

    pythonify:
flag to force instances of [`rdflib.term.Literal`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=Literal#rdflib.term.Identifier) to their Python literal representation

    returns:
a dictionary of serialized row bindings
        """
        bindings = {
            k: self.n3fy(v, pythonify=pythonify)
            for k, v in row_dict.items()
        }

        return bindings


    @classmethod
    def unbind_sparql (
        cls,
        query: str,
        bindings: dict,
        *,
        preamble: str = "",
        ) -> str:
        """
Substitute the _binding variables_ into the text of a SPARQL query,
to obviate the need for binding variables specified separately.
This can be helpful for debugging, or for some query engines that
may not have full SPARQL support yet.

    query:
text for the SPARQL query

    bindings:
variable bindings

    returns:
a string of the expanded SPARQL query
        """
        sparql_meta, sparql_body = re.split(r"\s*WHERE\s*\{", query, maxsplit=1)

        for var in sorted(bindings.keys(), key=lambda x: len(x), reverse=True):  # pylint: disable=W0108
            pattern = re.compile(r"(\?" + var + r")(\W)")
            bind_val = "<" + str(bindings[var]) + ">\\2"
            sparql_body = re.sub(pattern, bind_val, sparql_body)

        return "".join([preamble, sparql_meta, " WHERE {", sparql_body]).strip()


    def visualize (
        self,
        query: str,
        *,
        notebook: bool = False,
        ) -> pyvis.network.Network:
        """
Visualize the given SPARQL query as a [`pyvis.network.Network`](https://pyvis.readthedocs.io/en/latest/documentation.html#pyvis.network.Network)

    query:
input SPARQL query to be visualized

    notebook:
optional boolean flag, whether to initialize the PyVis graph to render within a notebook; defaults to `False`

    returns:
PyVis network object, to be rendered
        """
        return GPViz(query, self.kg._ns).visualize_query(notebook=notebook)  # pylint: disable=W0212



"""
Serialization-Deserialization for `KnowledgeGraph`
see license https://github.com/DerwenAI/kglab#license-and-copyright
"""

## Python standard libraries
import codecs
import datetime
import io
import json
import pathlib
import typing

### third-parties libraries
from icecream import ic  # type: ignore
import apeye.url
import chocolate  # type: ignore
import csvwlib  # type: ignore
import pandas as pd  # type: ignore

import rdflib  # type: ignore
import rdflib.plugin  # type: ignore
import rdflib.plugins.parsers.notation3 as rdf_n3  # type: ignore

## kglab - core classes
from .decorators import multifile
from .pkg_types import IOPathLike, PathLike
from .util import get_gpu_count, Mixin
from .version import _check_version


## pre-constructor set-up
_check_version()

if get_gpu_count() > 0:
    import cudf  # type: ignore  # pylint: disable=E0401


class SerdeMixin (Mixin):
    """
Provide serialization and deserialization methods for `KnowledgeGraph`:
* RDF
* JSONLD
* Parquet
* CSV
* ROAM
    """
    ######################################################################
    ## serialization
    ##
    ## Format and software agnostic: We are interested in technology
    ## that will not be a barrier to widespread use. We want to allow
    ## the export of the entity triples in a variety of formats, at
    ## the most basic level as a CSV, so that an analyst can load the
    ## results into their system of choice. By prioritizing
    ## non-proprietary, universal formats the results can be easily
    ## integrated into several of the existing tools for creating
    ## network graphs.

    _RDF_FORMAT: typing.Tuple = (
        # RDFXMLParser
        "application/rdf+xml",
        "xml",
        # N3Parser
        "text/n3",
        "n3",
        # TurtleParser
        "text/turtle",
        "turtle",
        "ttl",
        # NTParser
        "application/n-triples",
        "ntriples",
        "nt",
        "nt11",
        # NQuadsParser
        "application/n-quads",
        "nquads",
        # TriXParser
        "application/trix",
        "trix",
        # TrigParser
        "trig",
        # JsonLDParser
        "json-ld",
        "jelly",
    )

    _ERROR_ENCODE: str = "The text `encoding` value does not match anything in the Python codec registry"
    _ERROR_PATH: str = "The `path` file object must be a writable, bytes-like object"


    @classmethod
    def _check_format (
        cls,
        format: str,
        ) -> None:
        """
Semiprivate method to error-check that a `format` parameter corresponds to a known RDFlib serialization plugin;
otherwise this throws a `TypeError` exception
        """
        if format not in cls._RDF_FORMAT:
            try:
                rdflib.plugin.get(format, rdflib.serializer.Serializer)
            except Exception:
                raise TypeError("unknown format: {format}")


    @classmethod
    def _check_encoding (
        cls,
        encoding: str,
        ) -> None:
        """
Semiprivate method to error-check that an `encoding` parameter is within the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo);
otherwise this throws a `LookupError` exception
        """
        try:
            codecs.lookup(encoding)
        except LookupError:
            raise LookupError(cls._ERROR_ENCODE)


    @classmethod
    def _get_filename (
        cls,
        path: PathLike,
        ) -> typing.Optional[str]:
        """
Semiprivate method to extract a file name (str) for a file reference from a [`pathlib.Path`](https://docs.python.org/3/library/pathlib.html) or its subclasses

    returns:
a string as a file name or URL to a file reference
        """
        if not path:
            filename = None
        elif isinstance(path, apeye.url.URL):
            filename = str(path)
        elif isinstance(path, pathlib.Path):
            filename = path.as_posix()
        elif isinstance(path, str):
            filename = path
        else:
            raise TypeError(f"path variable not recognised {type(path)}")

        return filename


    @multifile()
    def load_rdf (
        self,
        path: IOPathLike,
        *,
        format: str = "ttl",
        base: typing.Optional[ str ] = None,
        **args: typing.Any,
        ) -> "KnowledgeGraph": # type: ignore
        """
Wrapper for [`rdflib.Graph.parse()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html#rdflib.graph.Graph.parse) which parses an RDF graph from the `path` source.
This traps some edge cases for the several source-ish parameters in RDFlib which had been overloaded.
Throws `TypeError` whenever a format parser plugin encounters a syntax error.

Note: this adds relations to an RDF graph, although it does not overwrite the existing RDF graph.

    path:
must be a file name (str) to a local file reference – possibly a glob with a wildcard; or a path object (not a URL) to a local file reference; or a [*readable, file-like object*](https://docs.python.org/3/glossary.html#term-file-object); otherwise this throws a `TypeError` exception

    format:
serialization format, defaults to Turtle triples; see `_RDF_FORMAT` for a list of default formats, which can be extended with plugins – excluding the `"json-ld"` format; otherwise this throws a `TypeError` exception

    base:
logical URI to use as the document base; if not specified, the document location gets used

    returns:
this `KnowledgeGraph` object – used for method chaining
        """
        # error checking for the `format` parameter
        if format == "json-ld":
            raise TypeError("Use the load_jsonld() method instead")

        self._check_format(format)

        # substitute the `KnowledgeGraph.base_uri` as the document base, if used
        if not base and self.base_uri:
            base = self.base_uri

        try:
            if hasattr(path, "read"):
                self._g.parse(  # type: ignore
                    path,  # type: ignore
                    format=format,
                    publicID=base,
                    **args,
                    )
            else:
                self._g.parse(  # type: ignore
                    self._get_filename(path),
                    format=format,
                    publicID=base,
                    **args,
                    )
        except rdf_n3.BadSyntax as e:
            ic(path)
            raise TypeError(str(e))

        return self


    def load_rdf_text (
        self,
        data: typing.AnyStr,
        *,
        format: str = "ttl",
        base: typing.Optional[ str ] = None,
        **args: typing.Any,
        ) -> "KnowledgeGraph": # type: ignore
        """
Wrapper for [`rdflib.Graph.parse()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html#rdflib.graph.Graph.parse) which parses an RDF graph from a text.
This traps some edge cases for the several source-ish parameters in RDFlib which had been overloaded.

Note: this adds relations to an RDF graph, it does not overwrite the existing RDF graph.

    data:
text representation of RDF graph data

    format:
serialization format, defaults to Turtle triples; see `_RDF_FORMAT` for a list of default formats, which can be extended with plugins – excluding the `"json-ld"` format; otherwise this throws a `TypeError` exception

    base:
logical URI to use as the document base; if not specified, the document location gets used

    returns:
this `KnowledgeGraph` object – used for method chaining
        """
        # error checking for the `format` parameter
        self._check_format(format)

        # substitute the `KnowledgeGraph.base_uri` as the document base, if used
        if not base and self.base_uri:
            base = self.base_uri

        self._g.parse( # type: ignore
            data=data,
            format=format,
            publicID=base,
            **args,
        )

        return self


    def save_rdf (
        self,
        path: IOPathLike,
        *,
        format: str = "ttl",
        base: typing.Optional[ str ] = None,
        encoding: str = "utf-8",
        **args: typing.Any,
        ) -> None:
        """
Wrapper for [`rdflib.Graph.serialize()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=serialize#rdflib.Graph.serialize) which serializes the RDF graph to the `path` destination.
This traps some edge cases for the `destination` parameter in RDFlib which had been overloaded.

    path:
must be a file name (str) or a path object (not a URL) to a local file reference; or a [*writable, bytes-like object*](https://docs.python.org/3/glossary.html#term-bytes-like-object); otherwise this throws a `TypeError` exception

    format:
serialization format, which defaults to Turtle triples; see `_RDF_FORMAT` for a list of default formats, which can be extended with plugins – excluding the `"json-ld"` format; otherwise this throws a `TypeError` exception

    base:
optional base set for the graph

    encoding:
optional text encoding value, defaults to `"utf-8"`, must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception
        """
        # error checking for the `format` parameter
        if format == "json-ld":
            raise TypeError("Use the save_jsonld() method instead")

        self._check_format(format)

        # error checking for the `encoding` parameter
        self._check_encoding(encoding)

        # substitute the `KnowledgeGraph.base_uri` base set for the graph, if used
        if not base and self.base_uri:
            base = self.base_uri

        # error checking for a file-like object `path` parameter
        if hasattr(path, "write"):
            if hasattr(path, "encoding"):
                raise TypeError(self._ERROR_PATH)

            try:
                self._g.serialize( # type: ignore
                    destination=path,  # type: ignore
                    format=format,
                    base=base,
                    encoding=encoding,
                    **args,
                    )
            except io.UnsupportedOperation:
                raise TypeError(self._ERROR_PATH)

        # otherwise write to a local file reference
        else:
            self._g.serialize( # type: ignore
                destination=self._get_filename(path),
                format=format,
                base=base,
                encoding=encoding,
                **args,
            )


    def save_rdf_text (
        self,
        *,
        format: str = "ttl",
        base: typing.Optional[ str ] = None,
        encoding: str = "utf-8",
        **args: typing.Any,
        ) -> str:
        """
Wrapper for [`rdflib.Graph.serialize()`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=serialize#rdflib.Graph.serialize) which serializes the RDF graph to a string.

    format:
serialization format, which defaults to Turtle triples; see `_RDF_FORMAT` for a list of default formats, which can be extended with plugins; otherwise this throws a `TypeError` exception

    base:
optional base set for the graph

    encoding:
optional text encoding value, defaults to `"utf-8"`, must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception

    returns:
text representing the RDF graph
        """
        # error checking for the `format` parameter
        self._check_format(format)

        # error checking for the `encoding` parameter
        self._check_encoding(encoding)

        # substitute the `KnowledgeGraph.base_uri` base set for the graph, if used
        if not base and self.base_uri:
            base = self.base_uri

        return self._g.serialize(  # type: ignore
            destination=None,
            format=format,
            base=base,
            encoding=encoding,
            **args,
        ).decode(encoding)


    @multifile()
    def load_jsonld (
        self,
        path: IOPathLike,
        *,
        encoding: str = "utf-8",
        **args: typing.Any,
        ) -> "KnowledgeGraph": # type: ignore
        """
Wrapper for [`rdflib-jsonld.parser.JsonLDParser.parse()`](https://github.com/RDFLib/rdflib-jsonld/blob/master/rdflib_jsonld/parser.py) which parses an RDF graph from a [JSON-LD](https://json-ld.org/) source.
This traps some edge cases for the several source-ish parameters in RDFlib which had been overloaded.

Note: this adds relations to an RDF graph, it does not overwrite the existing RDF graph.

    path:
must be a file name (str) to a local file reference – possibly a glob with a wildcard; or a path object (not a URL) to a local file reference; or a [*readable, file-like object*](https://docs.python.org/3/glossary.html#term-file-object)

    encoding:
optional text encoding value, which defaults to `"utf-8"`; must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception

    returns:
this `KnowledgeGraph` object – used for method chaining
        """
        # error checking for the `encoding` parameter
        self._check_encoding(encoding)

        # error checking for a file-like object `path` parameter
        if hasattr(path, "read"):
            f = path
        else:
            f = open(path, "r", encoding = encoding)  # type: ignore  # pylint: disable=R1732

        # load JSON from file (to verify format and trap exceptions at
        # this level) then dump to string – which is expected by the
        # JSON-LD plugin for RDFlib
        self._g.parse( # type: ignore
            data = json.dumps(json.load(f)),  # type: ignore
            format = "json-ld",
            encoding = encoding,
            **args,
        )

        return self


    def save_jsonld (
        self,
        path: IOPathLike,
        *,
        encoding: str = "utf-8",
        **args: typing.Any,
        ) -> None:
        """
Wrapper for [`rdflib-jsonld.serializer.JsonLDSerializer.serialize()`](https://github.com/RDFLib/rdflib-jsonld/blob/master/rdflib_jsonld/serializer.py) which serializes the RDF graph to the `path` destination as [JSON-LD](https://json-ld.org/).
This traps some edge cases for the `destination` parameter in RDFlib which had been overloaded.

    path:
must be a file name (str) or a path object (not a URL) to a local file reference; or a [*writable, bytes-like object*](https://docs.python.org/3/glossary.html#term-bytes-like-object); otherwise this throws a `TypeError` exception

    encoding:
optional text encoding value, which defaults to `"utf-8"`; must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception
        """
        # error checking for a file-like object `path` parameter
        if hasattr(path, "write"):
            if hasattr(path, "encoding"):
                raise TypeError(self._ERROR_PATH)

            f = path
        else:
            f = open(self._get_filename(path), "wb")  # type: ignore  # pylint: disable=R1732

        # error checking for the `encoding` parameter
        self._check_encoding(encoding)

        f.write( # type: ignore
            self._g.serialize(  # type: ignore
                format = "json-ld",
                context = self.get_context(),
                indent = 2,
                encoding = encoding,
                **args,
            )
        )


    _PARQUET_COL_NAMES: typing.List[str] = [
        "subject",
        "predicate",
        "object"
    ]

    @multifile()
    def load_parquet (
        self,
        path: IOPathLike,
        **kwargs: typing.Any,
        ) -> "KnowledgeGraph": # type: ignore
        """
Wrapper for [`pandas.read_parquet()`](https://pandas.pydata.org/docs/reference/api/pandas.read_parquet.html?highlight=read_parquet#pandas.read_parquet) which parses an RDF graph represented as a [Parquet](https://parquet.apache.org/) file, using the [`pyarrow`](https://arrow.apache.org/) engine.
Uses the [RAPIDS `cuDF` library](https://docs.rapids.ai/api/cudf/stable/) if GPUs are enabled.

To prepare for upcoming **kglab** features, **this is the preferred method for deserializing an RDF graph.**

Note: this adds relations to an RDF graph, it does not overwrite the existing RDF graph.

    path:
must be a file name (str) to a local file reference – possibly a glob with a wildcard; or a path object (not a URL) to a local file reference; or a [*readable, file-like object*](https://docs.python.org/3/glossary.html#term-file-object); a string could be a URL; valid URL schemes include `https`, `http`, `ftp`, `s3`, `gs`, `file`; a file URL can also be a path to a directory that contains multiple partitioned files, including a bucket in cloud storage – based on [`fsspec`](https://github.com/intake/filesystem_spec)

    returns:
this `KnowledgeGraph` object – used for method chaining
        """
        if self.use_gpus:
            df = cudf.read_parquet(  # pylint: disable=E0606
                path,
                **chocolate.filter_args(kwargs, pd.read_parquet)
            ).to_pandas()

        else:
            df = pd.read_parquet(
                path,
                **chocolate.filter_args(kwargs, pd.read_parquet)
            )

        df.apply(
            lambda row: self._g.parse(  # type: ignore
                data = f"{ row.iloc[0] } { row.iloc[1] } { row.iloc[2] } .",
                format = "ttl",
            ),
            axis = 1,
        )

        return self


    def save_parquet (
        self,
        path: IOPathLike,
        *,
        compression: str = "snappy",
        storage_options: typing.Optional[ dict ] = None,  # pylint: disable=W0613
        **kwargs: typing.Any,
        ) -> None:
        """
Wrapper for [`pandas.to_parquet()`](https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_parquet.html?highlight=to_parquet) which serializes an RDF graph to a [Parquet](https://parquet.apache.org/) file, using the [`pyarrow`](https://arrow.apache.org/) engine.
Uses the [RAPIDS `cuDF` library](https://docs.rapids.ai/api/cudf/stable/) if GPUs are enabled.

To prepare for upcoming **kglab** features, **this is the preferred method for serializing an RDF graph.**

    path:
must be a file name (str), path object to a local file reference, or a [*writable, bytes-like object*](https://docs.python.org/3/glossary.html#term-bytes-like-object); a string could be a URL; valid URL schemes include `https`, `http`, `ftp`, `s3`, `gs`, `file`; accessing cloud storage is based on [`fsspec`](https://github.com/intake/filesystem_spec)

    compression:
name of the compression algorithm to use; defaults to `"snappy"`; can also be `"gzip"`, `"brotli"`, or `None` for no compression

    storage_options:
extra options parsed by [`fsspec`](https://github.com/intake/filesystem_spec) for cloud storage access; **NOT USED** until `pandas` 1.2.x becomes stable across platforms and also RAPIDS provides support
        """
        rows_list: typing.List[dict] = [
            {
                self._PARQUET_COL_NAMES[0]: s.n3(),
                self._PARQUET_COL_NAMES[1]: p.n3(),
                self._PARQUET_COL_NAMES[2]: o.n3(),
            }
            for s, p, o in self._g # type: ignore
        ]

        if self.use_gpus:
            df = cudf.DataFrame(rows_list, columns=self._PARQUET_COL_NAMES)
        else:
            df = pd.DataFrame(rows_list, columns=self._PARQUET_COL_NAMES)

        df.to_parquet(
            path,
            compression=compression,
            #storage_options=storage_options,
            **chocolate.filter_args(kwargs, df.to_parquet),
        )


    def load_csv (
        self,
        url: str,
        ) -> "KnowledgeGraph": # type: ignore
        """
Wrapper for [`csvwlib`](https://github.com/DerwenAI/csvwlib) which parses a CSV file from the `path` source, then converts to RDF and merges into this RDF graph.

    url:
must be a URL represented as a string

    returns:
this `KnowledgeGraph` object – used for method chaining
        """
        new_rdf = csvwlib.CSVWConverter.to_rdf(
            url,
            mode="minimal",
            format="ttl",
        )
        return self.load_rdf_text(new_rdf)


    ######################################################################
    ## Roam Research integration

    def _walk_roam_graph (
        self,
        obj: dict,
        seen_uid: typing.Optional[ typing.Set[ str ] ] = None,
        ) -> str:
        """
Semiprivate method to traverse the Roam Research exported graph recursively, converting its objects into RDF representation.

    obj:
object to parse

    returns:
The `uid` identifier for the parsed object
        """
        if not seen_uid:
            seen_uid = set()

        roam_ns = str(self.get_ns("roam"))
        uid = obj["uid"]

        if uid not in seen_uid:
            seen_uid.add(uid)

            # create a node for this object
            node = rdflib.URIRef(roam_ns + uid)
            self.add(node, self.get_ns("rdf").type, self.get_ns("dct").Text)

            # represent title (complex object) or string description (simple object)
            if "title" in obj:
                descrip = obj["title"]
            elif "string" in obj:
                descrip = obj["string"]

            self.add(node, self.get_ns("skos").definition, rdflib.Literal(descrip))  # pylint: disable=E0606

            # represent the user who created/edited this object
            user_uid = obj[":edit/user"][":user/uid"]
            self.add(node, self.get_ns("dct").Creator, rdflib.URIRef(roam_ns + user_uid))

            # convert millisec timestamp to Unix epoch times (UTC) to datetime
            dt = datetime.datetime.utcfromtimestamp(round(obj["edit-time"] / 1000.0))
            self.add(node, self.get_ns("dct").Date, rdflib.Literal(dt.isoformat(), datatype = rdflib.XSD.dateTime))

            if "children" in obj:
                for child_obj in obj["children"]:
                    child_uid = self._walk_roam_graph(child_obj, seen_uid)
                    self.add(node, self.get_ns("dct").references, rdflib.URIRef(roam_ns + child_uid))

        return uid


    @multifile()
    def import_roam (
        self,
        path: IOPathLike,
        *,
        encoding: str = "utf-8",
        ) -> typing.List[str]:
        """
Import a graph in JSON that has been exported from the [Roam Research](https://roamresearch.com/) note-taking tool, then convert its objects and attributes into RDF representation.

For more details about the exported data from Roam Research, see:

  * <https://roamstack.com/roam-data-outside-roam/>
  * <https://nesslabs.com/roam-research-input-output>
  * <https://davidbieber.com/snippets/2020-04-25-roam-json-export/>

Note: this adds relations to an RDF graph, it does not overwrite the existing RDF graph.

    path:
must be a file name (str) to a local file reference – possibly a glob with a wildcard; or a path object (not a URL) to a local file reference; or a [*readable, file-like object*](https://docs.python.org/3/glossary.html#term-file-object)

    encoding:
optional text encoding value, which defaults to `"utf-8"`; must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception

    returns:
a list of identifiers for the top-level nodes added from the Roam Research graph
        """
        # error checking for the `encoding` parameter
        self._check_encoding(encoding)

        # error checking for a file-like object `path` parameter
        if hasattr(path, "read"):
            f = path
        else:
            f = open(path, "r", encoding = encoding)  # type: ignore  # pylint: disable=R1732

        # add a `roam:` prefix for a pseudo-namespace to use here,
        # which applications may need to parameterize later?
        self.add_ns("roam", "https://roamresearch.com/ns/")

        uid_list: typing.List[str] = [
            self._walk_roam_graph(obj)
            for obj in json.load(f)  # type: ignore
            ]

        return uid_list



"""
Working with `SubgraphMatrix` as vectorized representation.
Additions to functionalities present in `subg.py`.
Integrate `scipy` and `scikit-learn` functionalities.

see license https://github.com/DerwenAI/kglab#license-and-copyright
"""

import typing

import networkx as nx

from .util import Mixin


class AlgebraMixin (Mixin):
    """
Provides methods to work with graph algebra using `SubgraphMatrix` data.

NOTE: provide optional Oxigraph support for fast in-memory computation
    """
    nx_graph: typing.Optional[nx.DiGraph] = None


    def to_undirected (self):
        """
Return the undirected adjancency matrix of the directed graph.

        returns:
`numpy.array`: the array representation in `numpy` standard
        """
        return nx.to_numpy_array(self.nx_graph.to_undirected())


    def to_adjacency (self):
        """
Return adjacency (dense) matrix for the KG.
[Relevant NetworkX interface](https://networkx.org/documentation/stable/reference/convert.html#id2)

        returns:
`numpy.array`: the array representation in `numpy` standard
        """
        self.check_attributes()
        return nx.to_numpy_array(self.nx_graph)


    def to_incidence (self):
        """
Return incidence (dense) matrix for the KG.
[Relevant scipy docs](https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.csr_matrix.html)

        returns:
`numpy.array`: the array representation in `numpy` standard
        """
        self.check_attributes()
        return nx.incidence_matrix(self.nx_graph).toarray()


    def to_laplacian (self):
        """
Return Laplacian matrix for the KG. Graph is turned into undirected.
[docs](https://networkx.org/documentation/stable/reference/generated/networkx.linalg.laplacianmatrix.laplacian_matrix.html).
Lapliacian is also known as vertices degrees matrix.

        returns:
`numpy.array`: the array representation in `numpy` standard
        """
        self.check_attributes()
        return nx.laplacian_matrix(self.nx_graph.to_undirected()).toarray()


    def to_scipy_sparse (self):
        """
Return graph in CSR format (optimized for matrix-matrix operations).

        returns:
SciPy sparse matrix: Graph adjacency matrix.
        """
        self.check_attributes()
        return nx.to_scipy_sparse_array(self.nx_graph)


    def to_zarr (self, func: str) -> "zarr.core.Array":  # type: ignore
        """
Return value of a function in `kglab.algebra` in [Zarr format](https://pypi.org/project/zarr/).

       Args:
func str: a name of one functions in `kglab.algebra`: `to_undirect`, `to_adjacency`, `to_incidence`, `to_laplacian`

       returns:
zzarr.core.Array: values of requested `func`
        """
        self.check_attributes()  # type: ignore

        try:
            import zarr  # type: ignore
        except ImportError:
            raise ImportError("To use Zarr you need to install kglab with the required extra package: pip install kglab[with-zarr]")

        data = getattr(self, func)()
        array = zarr.create(data.shape, chunks = True)
        array[:] = data

        return array



"""
Mixin class to provide SHACL- OWL- and RDF-related functionalities for `KnowledgeGraph`

see license https://github.com/DerwenAI/kglab#license-and-copyright
"""

## Python standard libraries
import typing

### third-parties libraries
import chocolate  # type: ignore
import owlrl  # type: ignore
import pyshacl  # type: ignore

## kglab - core classes
from .pkg_types import GraphLike
from .util import Mixin


class ShaclOwlRdfSkosMixin (Mixin):
    """
Provide methods for SHACL- OWL- and RDF-related operations.
    """
    _g: typing.Optional[GraphLike]

    ######################################################################
    ## SHACL validation

    def validate (
        self,
        *,
        shacl_graph: typing.Optional[typing.Union[GraphLike, typing.AnyStr]] = None,
        shacl_graph_format: typing.Optional[str] = None,
        ont_graph: typing.Optional[typing.Union[GraphLike, typing.AnyStr]] = None,
        ont_graph_format: typing.Optional[str] = None,
        advanced: typing.Optional[bool] = False,
        inference: typing.Optional[str] = None,
        inplace:typing.Optional[bool] = True,
        abort_on_first: typing.Optional[bool] = None,
        **kwargs: typing.Any,
        ) -> typing.Tuple[bool, "KnowledgeGraph", str]: # type: ignore
        """
Wrapper for [`pyshacl.validate()`](https://github.com/RDFLib/pySHACL) for validating the RDF graph using rules expressed in the [SHACL](https://www.w3.org/TR/shacl/) (Shapes Constraint Language).

    shacl_graph:
text representation, file path, or URL of the SHACL *shapes graph* to use in validation

    shacl_graph_format:
RDF format, if the `shacl_graph` parameter is a text representation of the *shapes graph*

    ont_graph:
text representation, file path, or URL of an optional, extra ontology to mix into the RDF graph

    ont_graph_format
RDF format, if the `ont_graph` parameter is a text representation of the extra ontology

    advanced:
enable advanced SHACL features

    inference:
prior to validation, run OWL2 RL profile-based expansion of the RDF graph based on [OWL-RL](https://github.com/RDFLib/OWL-RL); values: `"rdfs"`, `"owlrl"`, `"both"`, `None`

    inplace:
when enabled, do not clone the RDF graph prior to inference/expansion, just manipulate it in-place

    abort_on_first:
abort validation on the first error

    returns:
a tuple of `conforms` (RDF graph passes the validation rules) + `report_graph` (report as a `KnowledgeGraph` object) + `report_text` (report formatted as text)
        """
        conforms, report_graph_data, report_text = pyshacl.validate(
            self._g,  # type: ignore
            shacl_graph=shacl_graph,
            shacl_graph_format=shacl_graph_format,
            ont_graph=ont_graph,
            ont_graph_format=ont_graph_format,
            advanced=advanced,
            inference=inference,
            inplace=inplace,
            abort_on_first=abort_on_first,
            serialize_report_graph="ttl",
            **chocolate.filter_args(kwargs, pyshacl.validate),
            )

        g = self.build_blank_graph()

        g.parse(
            data=report_graph_data,
            format="ttl",
            encoding="utf-8"
        )

        report_graph = self.graph_factory(
            name="SHACL report graph",
            graph=g,
        )

        return conforms, report_graph, report_text


    ######################################################################
    ## OWL RL inference
    ## adapted from <https://wiki.uib.no/info216/index.php/Python_Examples>

    def infer_owlrl_closure (
        self
        ) -> None:
        """
Infer deductive closure for [OWL 2 RL semantics](https://www.w3.org/TR/owl2-profiles/#Reasoning_in_OWL_2_RL_and_RDF_Graphs_using_Rules) based on [OWL-RL](https://owl-rl.readthedocs.io/en/latest/stubs/owlrl.html#module-owlrl)

See <https://wiki.uib.no/info216/index.php/Python_Examples#RDFS_inference_with_RDFLib>
        """
        owl = owlrl.OWLRL_Semantics(self._g, False, False, False)
        owl.closure()
        owl.flush_stored_triples()


    def infer_rdfs_closure (
        self
        ) -> None:
        """
Infer deductive closure for [RDFS semantics](https://www.w3.org/TR/rdf-mt/) based on [OWL-RL](https://owl-rl.readthedocs.io/en/latest/stubs/owlrl.html#module-owlrl)

See <https://wiki.uib.no/info216/index.php/Python_Examples#RDFS_inference_with_RDFLib>
        """
        rdfs = owlrl.RDFSClosure.RDFS_Semantics(self._g, False, False, False)
        rdfs.closure()
        rdfs.flush_stored_triples()


    def infer_rdfs_properties (
        self
        ) -> None:
        """
Perform RDFS sub-property inference, adding super-properties where sub-properties have been used.

Adapted from [`skosify`](https://github.com/NatLibFi/Skosify) which wasn't being updated regularly.
        """
        _rdfs = self.get_ns("rdfs")

        # determine sub-property mappings
        # key: property val: set([superprop1, superprop2..])
        super_props: typing.Dict[typing.Any, typing.Any] = {}

        for s, o in self._g.subject_objects(_rdfs.subPropertyOf): # type: ignore
            super_props.setdefault(s, set())

            for sub_prop in self._g.transitive_objects(s, _rdfs.subPropertyOf): # type: ignore
                if sub_prop != s:
                    super_props[s].add(sub_prop)

        # add super-property relationships
        for p, sup_prop_list in super_props.items():
            for s, o in self._g.subject_objects(p): # type: ignore
                for sup_prop in sup_prop_list:
                    self.add(s, sup_prop, o)


    def infer_rdfs_classes (
        self
        ) -> None:
        """
Perform RDFS subclass inference, marking all resources having a subclass type with their superclass.

Adapted from [`skosify`](https://github.com/NatLibFi/Skosify) which wasn't being updated regularly.
        """
        _rdfs = self.get_ns("rdfs")

        # determine subclass mappings
        # key: class val: set([superclass1, superclass2..])
        super_classes: typing.Dict[typing.Any, typing.Any] = {}

        for s, _ in self._g.subject_objects(_rdfs.subClassOf): # type: ignore
            super_classes.setdefault(s, set())

            for sup_class in self._g.transitive_objects(s, _rdfs.subClassOf): # type: ignore
                if sup_class != s:
                    super_classes[s].add(sup_class)

        # set the superclass type information for subclass instances
        for s, sup_class_list in super_classes.items():
            for sub_inst in self._g.subjects(self.get_ns("rdf").type, s): # type: ignore
                for sup_class in sup_class_list:
                    self.add(sub_inst, self.get_ns("rdf").type, sup_class)


    ######################################################################
    ## SKOS inference
    ## adapted from `skosify` https://github.com/NatLibFi/Skosify
    ## it wasn't being updated regularly, but may be integrated again

    def infer_skos_related (
        self
        ) -> None:
        """
Infer OWL symmetry (both directions) for `skos:related`
[(*S23*)](https://www.w3.org/TR/skos-reference/#S23)

Adapted from [`skosify`](https://github.com/NatLibFi/Skosify) which wasn't being updated regularly.
        """
        _skos = self.get_ns("skos")

        for s, o in self._g.subject_objects(_skos.related): # type: ignore
            self.add(o, _skos.related, s)


    def infer_skos_concept (
        self
        ) -> None:
        """
Infer `skos:topConceptOf` as a sub-property of `skos:inScheme`
[(*S7*)](https://www.w3.org/TR/skos-reference/#S7)

Infer `skos:topConceptOf` as `owl:inverseOf` the property `skos:hasTopConcept`
[(*S8*)](https://www.w3.org/TR/skos-reference/#S8)

Adapted from [`skosify`](https://github.com/NatLibFi/Skosify) which wasn't being updated regularly.
        """
        _skos = self.get_ns("skos")

        for s, o in self._g.subject_objects(_skos.hasTopConcept): # type: ignore
            self.add(o, _skos.topConceptOf, s)

        for s, o in self._g.subject_objects(_skos.topConceptOf): # type: ignore
            self.add(o, _skos.hasTopConcept, s)

        for s, o in self._g.subject_objects(_skos.topConceptOf): # type: ignore
            self.add(s, _skos.inScheme, o)


    def infer_skos_hierarchical (
        self,
        *,
        narrower: bool = True,
        ) -> None:
        """
Infer `skos:narrower` as `owl:inverseOf` the property `skos:broader`; although only keep `skos:narrower` on request
[(*S25*)](https://www.w3.org/TR/skos-reference/#S25)

Adapted from [`skosify`](https://github.com/NatLibFi/Skosify) which wasn't being updated regularly.

    narrower:
if false, `skos:narrower` will be removed instead of added
        """
        _skos = self.get_ns("skos")

        if narrower:
            for s, o in self._g.subject_objects(_skos.broader): # type: ignore
                self.add(o, _skos.narrower, s)

        for s, o in self._g.subject_objects(_skos.narrower): # type: ignore
            self.add(o, _skos.broader, s)

            if not narrower:
                self.remove(s, _skos.narrower, o)


    def infer_skos_transitive (
        self,
        *,
        narrower: bool = True,
        ) -> None:
        """
Infer transitive closure,
`skos:broader` as a sub-property of `skos:broaderTransitive`, and `skos:narrower` as a sub-property of `skos:narrowerTransitive`
[(*S22*)](https://www.w3.org/TR/skos-reference/#S22)

Infer `skos:broaderTransitive` and `skos:narrowerTransitive` (on request only) as instances of `owl:TransitiveProperty`
[(*S24*)](https://www.w3.org/TR/skos-reference/#S24)

Adapted from [`skosify`](https://github.com/NatLibFi/Skosify) which wasn't being updated regularly.

    narrower:
also infer transitive closure for `skos:narrowerTransitive`
        """
        _skos = self.get_ns("skos")

        for concept in self._g.subjects(self.get_ns("rdf").type, _skos.Concept): # type: ignore
            for broader_concept in self._g.transitive_objects(concept, _skos.broader): # type: ignore
                if broader_concept != concept:
                    self.add(concept, _skos.broaderTransitive, broader_concept)

                    if narrower:
                        self.add(broader_concept, _skos.narrowerTransitive, concept)


    def infer_skos_symmetric_mappings (
        self,
        *,
        related: bool = True,
        ) -> None:
        """
Infer symmetric mapping properties (`skos:relatedMatch`, `skos:closeMatch`, `skos:exactMatch`) as instances of `owl:SymmetricProperty`
[(*S44*)](https://www.w3.org/TR/skos-reference/#S44)

Adapted from [`skosify`](https://github.com/NatLibFi/Skosify) which wasn't being updated regularly.

    related:
infer the `skos:related` super-property for all `skos:relatedMatch` relations
[(*S41*)](https://www.w3.org/TR/skos-reference/#S41)
        """
        _skos = self.get_ns("skos")

        for s, o in self._g.subject_objects(_skos.relatedMatch): # type: ignore
            self.add(o, _skos.relatedMatch, s)

            if related:
                self.add(s, _skos.related, o)
                self.add(o, _skos.related, s)

        for s, o in self._g.subject_objects(_skos.closeMatch): # type: ignore
            self.add(o, _skos.closeMatch, s)

        for s, o in self._g.subject_objects(_skos.exactMatch): # type: ignore
            self.add(o, _skos.exactMatch, s)


    def infer_skos_hierarchical_mappings (
        self,
        *,
        narrower: bool = True,
        ) -> None:
        """
Infer `skos:narrowMatch` as `owl:inverseOf` the property `skos:broadMatch`
[(*S43*)](https://www.w3.org/TR/skos-reference/#S43)

Infer the `skos:related` super-property for all `skos:relatedMatch` relations
[(*S41*)](https://www.w3.org/TR/skos-reference/#S41)

Adapted from [`skosify`](https://github.com/NatLibFi/Skosify) which wasn't being updated regularly.

        narrower:
if false, `skos:narrowMatch` will be removed instead of added
        """
        _skos = self.get_ns("skos")

        for s, o in self._g.subject_objects(_skos.broadMatch): # type: ignore
            self.add(s, _skos.broader, o)

            if narrower:
                self.add(o, _skos.narrowMatch, s)
                self.add(o, _skos.narrower, s)

        for s, o in self._g.subject_objects(_skos.narrowMatch): # type: ignore
            self.add(o, _skos.broadMatch, s)
            self.add(o, _skos.broader, s)

            if narrower:
                self.add(s, _skos.narrower, o)
            else:
                self.remove(s, _skos.narrowMatch, o)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/kglab#license-and-copyright

"""
Utility functions used within `kglab` and related applications.
"""

import math
import typing

import numpy as np  # type: ignore  # pylint: disable=E0401
import pandas as pd  # type: ignore  # pylint: disable=E0401

from .pkg_types import GraphLike

GPU_COUNT: int = 0


def get_gpu_count () -> int:
    """
Special handling for detecting GPU availability: an approach
recommended by the NVidia RAPIDS engineering team, since `nvml`
bindings are difficult for Python libraries to keep updated.

    returns:
count of available GPUs, where `0` means none or disabled.
    """
    global GPU_COUNT  # pylint: disable=W0603

    if GPU_COUNT < 0:
        return 0

    try:
        import pynvml  # type: ignore  # pylint: disable=E0401
        pynvml.nvmlInit()

        GPU_COUNT = pynvml.nvmlDeviceGetCount()
    except Exception: # pylint: disable=W0703
        GPU_COUNT = -1

    return GPU_COUNT


if get_gpu_count() > 0:
    try:
        import cudf  # type: ignore # pylint: disable=E0401
    except Exception as gpu_e: # pylint: disable=W0703
        # turn off GPU usage
        #print(gpu_e)
        GPU_COUNT = -1


def calc_quantile_bins (
    num_rows: int
    ) -> np.ndarray:
    """
Calculate the bins to use for a quantile stripe, using [`numpy.linspace`](https://numpy.org/doc/stable/reference/generated/numpy.linspace.html)

    num_rows:
number of rows in the target dataframe

    returns:
the calculated bins, as a [`numpy.ndarray`](https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html)
    """
    granularity = max(round(math.log(num_rows) * 4), 1)
    return np.linspace(0, 1, num=granularity, endpoint=True)


def stripe_column (
    values: list,
    bins: int,
    *,
    use_gpus: bool = False,
    ) -> np.ndarray:
    """
Stripe a column in a dataframe, by interpolating quantiles into a set of discrete indexes.

    values:
list of values to stripe

    bins:
quantile bins; see [`calc_quantile_bins()`](#calc_quantile_bins-function)

    use_gpus:
optionally, use the NVidia GPU devices with the [RAPIDS libraries](https://rapids.ai/) if these libraries have been installed and the devices are available; defaults to `False`

    returns:
the striped column values, as a [`numpy.ndarray`](https://numpy.org/doc/stable/reference/generated/numpy.ndarray.html); uses the [RAPIDS `cuDF` library](https://docs.rapids.ai/api/cudf/stable/) if GPUs are enabled
    """
    if use_gpus:
        s = cudf.Series(values)  # pylint: disable=E0606
    else:
        s = pd.Series(values)

    q = s.quantile(bins, interpolation="nearest")

    try:
        stripe = np.digitize(values, q) - 1
        return stripe
    except ValueError as e:
        # should never happen?
        print("ValueError:", str(e), values, s, q, bins)
        raise


def root_mean_square (
    values: list
    ) -> float:
    """
Calculate the [*root mean square*](https://mathworld.wolfram.com/Root-Mean-Square.html) of the values in the given list.

    values:
list of values to use in the RMS calculation

    returns:
RMS metric as a float
    """
    s = sum(map(lambda x: float(x)**2.0, values))
    n = float(len(values))
    return math.sqrt(s / n)

class Mixin:
    """Base mixin, Provide `mypy` stubs for common methods and properties"""
    _g: typing.Optional[GraphLike]
    get_ns: typing.Callable
    add_ns: typing.Callable
    _ns: typing.Dict
    add: typing.Callable
    base_uri: typing.Optional[str]
    parse: typing.Callable
    get_context: typing.Callable
    use_gpus: bool
    serialize: typing.Callable
    build_blank_graph: typing.Callable
    graph_factory: typing.Callable
    remove: typing.Callable



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Decorators:

  * `@multifile` handles specifying multiple paths in serialization methods.
"""

from copy import deepcopy
from functools import wraps
from glob import glob
import inspect
import pathlib
import typing

from icecream import ic  #  type: ignore # pylint: disable=W0611,E0401
import apeye.url


def _test_path (
    path: typing.Union[ pathlib.Path, apeye.url.URL ]
    ) -> bool:
    """
Semi-private function to test whether the given path parameter is an
instance of `pathlib.Path`, `apeye.url.URL`, or a related class.

    path:
parameter to be tested

    returns:
test condition
    """
    return isinstance(path, (pathlib.Path, apeye.url.URL))


def multifile (
    param_name: str = "path",
    ) -> typing.Any:
    """
Creates a wrapper around a read function to read multiple files, given
a glob pattern with at least one wildcard in the path.

    param_name:
parameter to be overloaded

    returns:
constructed decorator
    """

    def decorator (
        f: typing.Any,
        ) -> typing.Any:
        sig = inspect.signature(f)

        if param_name not in sig.parameters:
            raise ValueError(
                f"Reader function {f.__name__} has no parameter '{param_name}'"
            )

        @wraps(f)
        def wrapper (  # pylint: disable=R1710
            *args: typing.Any,
            **kwargs: typing.Any,
            ) -> typing.Any:
            bound_arguments = sig.bind(*args, **kwargs)
            bound_arguments.apply_defaults()

            path = bound_arguments.arguments[param_name]
            path_list: typing.List = []

            # if `*` is not in the path then simply let the default
            # function handle it
            if isinstance(path, str):
                if "*" not in path:
                    return f(*args, **kwargs)

                # initialize the path list with a parsed glob
                path_list = glob(path)

            # handle single Path objects by the default function
            elif _test_path(path):
                return f(*args, **kwargs)

            # handle a list of Path objects
            elif isinstance(path, list):
                for p in path:
                    if _test_path(p):
                        path_list.append(p)
                    else:
                        raise ValueError(f"Invalid path: {p}")
            else:
                raise ValueError(
                    f"{path} is not a valid string, Path, or list of Paths"
                )

            # no files were found in the given pattern so raise error
            if len(path_list) == 0:
                raise ValueError(f"No files found in given path list: {path}")

            # iterate through each path in the glob
            for p in path_list:
                # set the path variable
                bound_arguments.arguments[param_name] = str(p)

                # call the underlying reader function
                p_result = f(*bound_arguments.args, **deepcopy(bound_arguments.kwargs))  # pylint: disable=W0612

        return wrapper

    return decorator



"""
see license https://github.com/DerwenAI/kglab#license-and-copyright
"""
from .kglab import KnowledgeGraph

from .graph import NodeRef, PropertyStore

from .subg import Subgraph, SubgraphMatrix, SubgraphTensor

from .topo import Measure, Simplex0, Simplex1

from .pkg_types import RDF_Triple, RDF_Node, SPARQL_Bindings, \
    PathLike, IOPathLike, \
    NodeLike, GraphLike, \
    Census_Item, Census_Dyad_Tally

from .external_import import import_from_neo4j

from .gpviz import GPViz

from .util import get_gpu_count, calc_quantile_bins, stripe_column, root_mean_square

from .version import MIN_PY_VERSION, _versify, _check_version, __version__


# deprecated:
#from .esp import ShapeFactory, EvoShape, EvoShapeNode, EvoShapeEdge, \
#    Leaderboard, SerializedEvoShape, EvoShapeBoard, EvoShapeDistance

#from .srl import PSLModel



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Storage plugin for RDFlib

see license https://github.com/DerwenAI/kglab#license-and-copyright
"""

from dataclasses import dataclass
import inspect
import typing

from cryptography.hazmat.primitives import hashes  # type: ignore  # pylint: disable=E0401
from icecream import ic  # type: ignore  # pylint: disable=E0401,W0611
import chocolate  # type: ignore  # pylint: disable=E0401
import numpy as np  # type: ignore  # pylint: disable=E0401

from rdflib.plugins.sparql.sparql import Query, Update
from rdflib.query import Result
from rdflib.store import Store  # type: ignore # pylint: disable=E0401
from rdflib.term import Identifier
import rdflib  # type: ignore  # pylint: disable=E0401


@dataclass(frozen=True)
class NodeRef:
    """
Represent a reference to a Node within this store.
    """
    id: int
    node_name: str
    node_id: int


class PropertyStore (Store):
    """
A subclass of `rdflib.Store` to use as a plugin, integrating the W3C stack.
    """

    def __init__ (
        self,
        configuration: typing.Optional[str] = None,
        identifier: typing.Optional[str] = None,
        ) -> None:
        """
Instance constructor.
        """
        if configuration is not None:
            ic(configuration)

        super().__init__(configuration)

        self.identifier = identifier
        self.digest: typing.Optional[ hashes.Hash ] = None

        self.__namespace: dict = {}
        self.__prefix: dict = {}

        self._tuples: list = []
        self._node_names: np.array = np.empty(shape=[0, 1], dtype=object)  # type: ignore
        self._rel_names: np.array = np.empty(shape=[0, 1], dtype=object)  # type: ignore


######################################################################
## rdflib.Store implementation

    @classmethod
    def get_lpg (
        cls,
        graph: rdflib.Graph,
        ) -> "PropertyStore":
        """
An accessor method to extract the PropertyGraph from an RDF graph,
which is a private member of rdflib.Graph.
        """
        return graph._Graph__store  # type: ignore # pylint: disable=W0212


    def get_node_id (
        self,
        node_name: str,
        ) -> int:
        """
An accessor method to map from the unique name of a node to its
`node_id` index.
        """
        rows = np.where(self._node_names == node_name)

        if len(rows[0]) < 1:
            self._node_names = np.append(self._node_names, node_name)
            node_id = len(self._node_names) - 1
        else:
            node_id = rows[0][0]

        return node_id


    def get_node_name (
        self,
        node_id: int,
        ) -> str:
        """
An accessor method to map from the `node_id` index of a node to its
unique name.
        """
        return self._node_names[node_id]  # type: ignore


    def get_rel_id (
        self,
        rel_name: str,
        ) -> int:
        """
An accessor method to map from a unique name of a relation to its
`rel_id` index.
        """
        rows = np.where(self._rel_names == rel_name)

        if len(rows[0]) < 1:
            self._rel_names = np.append(self._rel_names, rel_name)
            rel_id = len(self._rel_names) - 1
        else:
            rel_id = rows[0][0]

        return rel_id


    def get_rel_name (
        self,
        rel_id: int,
        ) -> str:
        """
An accessor method to map from the `rel_id` index of a relation to its
unique name.
        """
        return self._rel_names[rel_id]  # type: ignore


    def build_tuple (
        self,
        s,
        p,
        o,
        context,
        ) -> typing.Tuple:
        """
Compose a tuple from the inputs supplied by `RDFlib`.
        """
        if context is None:
            c = None
        else:
            c = str(context.identifier)  # type: ignore

        src_id = self.get_node_id(str(s))
        rel_id = self.get_rel_id(str(p))

        if isinstance(o, rdflib.term.Literal):
            _tuple = ( src_id, rel_id, str(o), True, c)
        else:
            dst_id = self.get_node_id(str(o))
            _tuple = ( src_id, rel_id, dst_id, False, c)  # type: ignore

        return _tuple


    def _find (
        self,
        _tuple: typing.Tuple,
        ) -> int:
        """
Locate the given tuple in the data, returning `-1` if not found.
        """
        try:
            idx = self._tuples.index(_tuple)
        except ValueError as ex:  # pylint: disable=W0612
            # triple does not exist
            idx = -1

        return idx


    def add (  # type: ignore # pylint: disable=W0221
        self,
        triple: typing.Tuple,
        context: typing.Optional[ rdflib.term.URIRef ] = None,  # pylint: disable=W0613
        *,
        quoted: bool = False,  # pylint: disable=W0613
        ) -> None:
        """
Adds the given statement to a specific context or to the model.

The quoted argument is interpreted by formula-aware stores to indicate
this statement is quoted/hypothetical.

It should be an error to not specify a context and have the quoted
argument be `True`.

It should also be an error for the quoted argument to be `True` when
the store is not formula-aware.
        """
        s, p, o = triple  # pylint: disable=W0612
        _tuple = self.build_tuple(str(s), str(p), o, context)
        idx = self._find(_tuple)

        if idx < 0:
            self._tuples.append(_tuple)

            # update digest
            if self.digest is not None:
                self.digest.update(inspect.currentframe().f_code.co_name.encode("utf-8"))  # type: ignore
                self.digest.update(str(s).encode("utf-8"))
                self.digest.update(str(p).encode("utf-8"))
                self.digest.update(str(o).encode("utf-8"))

                c = _tuple[4]

                if c is not None:
                    self.digest.update(c.encode("utf-8"))


    def remove (  # type: ignore # pylint: disable=W0221
        self,
        triple_pattern: typing.Tuple,
        *,
        context: typing.Optional[ rdflib.term.URIRef ] = None,  # pylint: disable=W0613
        ) -> None:
        """
Remove the set of triples matching the pattern from the store.
        """
        s, p, o = triple_pattern  # pylint: disable=W0612
        _tuple = self.build_tuple(str(s), str(p), o, context)
        idx = self._find(_tuple)

        if idx >= 0:
            del self._tuples[idx]

            # update digest
            if self.digest is not None:
                self.digest.update(inspect.currentframe().f_code.co_name.encode("utf-8"))  # type: ignore
                self.digest.update(str(s).encode("utf-8"))
                self.digest.update(str(p).encode("utf-8"))
                self.digest.update(str(o).encode("utf-8"))

                c = _tuple[4]

                if c is not None:
                    self.digest.update(c.encode("utf-8"))


    def triples (  # type: ignore # pylint: disable=W0221
        self,
        triple_pattern: typing.Tuple,
        *,
        context: typing.Optional[ rdflib.term.URIRef ] = None,  # pylint: disable=W0613
        ) -> typing.Generator:
        """
A generator over all the triples matching the pattern.

    triple_pattern:
Can include any objects for used for comparing against nodes in the store, for example, REGEXTerm, URIRef, Literal, BNode, Variable, Graph, QuotedGraph, Date? DateRange?

    context:
A conjunctive query can be indicated by either providing a value of None, or a specific context can be queries by passing a Graph instance (if store is context aware).
        """
        s, p, o = triple_pattern  # pylint: disable=W0612

        if s is not None:
            s = str(s)

        if p is not None:
            p = str(p)

        if o is not None:
            o = str(o)

        if context is None:
            c = None
        else:
            c = str(context.identifier)  # type: ignore

        #_tuple = ( s, p, o, o_lit, c, )

        for src, rel, dst, o_lit, ctx in self._tuples:  # pylint: disable=R1702
            if (s is None) or (s == src):
                if (p is None) or (p == rel):
                    if (o is None) or (o == dst):
                        if (c is None) or (c == ctx):
                            if o_lit:
                                dst_ref: typing.Any = rdflib.term.Literal(dst)
                            else:
                                dst_ref = self.get_node_name(dst)

                            triple_result = (
                                rdflib.term.URIRef(self.get_node_name(src)),
                                rdflib.term.URIRef(self.get_rel_name(rel)),
                                dst_ref,
                            )

                            yield triple_result, self.__contexts()


    def __len__ (  # type: ignore # pylint: disable=W0221,W0222
        self,
        *,
        context: typing.Optional[ rdflib.term.URIRef ] = None,  # pylint: disable=W0613
        ) -> int:
        """
Number of statements in the store. This should only account for
non-quoted (asserted) statements if the context is not specified,
otherwise it should return the number of statements in the formula or
context given.

    context:
a graph instance to query or None
        """
        if context is None:
            return len(self._tuples)

        c = str(context.identifier)  # type: ignore
        count = 0

        for _, _, _, _, ctx in self._tuples:
            if c == ctx:
                count += 1

        return count


    def __contexts (
        self
        ) -> typing.Iterable:
        """
Returns the set of contexts
        """
        return { ctx for _, _, _, _, ctx in self._tuples }


    def bind (  # pylint: disable=W0221
        self,
        prefix: str,
        namespace: rdflib.term.URIRef,
        override: bool = True,
        ) -> None:
        """
Foo.
        """
        self.__prefix[namespace] = prefix
        self.__namespace[prefix] = namespace


    def namespace (
        self,
        prefix: str,
        ) -> rdflib.term.URIRef:
        """
Bar.
        """
        return self.__namespace.get(prefix, None)  # type: ignore


    def prefix (
        self,
        namespace: str,
        ) -> str:
        """
Bar.
        """
        return self.__prefix.get(namespace, None)  # type: ignore


    def namespaces (
        self
        ) -> typing.Iterator[ tuple[ str, rdflib.term.URIRef ] ]:
        """
Bar.
        """
        for prefix, namespace in self.__namespace.items():
            yield prefix, namespace


    def query (  # pylint: disable=W0235
        self,
        query: Query | str,
        initNs: typing.Mapping[ str, typing.Any ],
        initBindings: typing.Mapping[ str, Identifier ],
        queryGraph: str,
        **kwargs: typing.Any,
        ) -> Result:
        """
queryGraph is None, a URIRef or '__UNION__'

If None the graph is specified in the query-string/object

If URIRef it specifies the graph to query,

If  '__UNION__' the union of all named graphs should be queried

This is used by Datasets
Values other than None obviously only makes sense for context-aware stores.)
        """
        return super().query(
            query,
            initNs,
            initBindings,
            queryGraph,
            **chocolate.filter_args(kwargs, super().query),
        )


    def update (  # pylint: disable=W0235
        self,
        update: Update | str,
        initNs: typing.Mapping[ str, typing.Any ],
        initBindings: typing.Mapping[ str, Identifier ],
        queryGraph: str,
        **kwargs: typing.Any,
        ) -> None:
        """
queryGraph is None, a URIRef or '__UNION__'

If None the graph is specified in the query-string/object

If URIRef it specifies the graph to query,

If  '__UNION__' the union of all named graphs should be queried

This is used by Datasets
Values other than None obviously only makes sense for context-aware stores.)
        """
        super().update(
            update,
            initNs,
            initBindings,
            queryGraph,
            **chocolate.filter_args(kwargs, super().update),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/kglab#license-and-copyright

"""
Graph topology.
"""

from collections import defaultdict
import typing

import pandas as pd  # type: ignore  # pylint: disable=E0401
import rdflib  # type: ignore  # pylint: disable=E0401

from .kglab import KnowledgeGraph
from .pkg_types import Census_Item, Census_Dyad_Tally


class Simplex0:
    """
Count the distribution of a class of items in an RDF graph.
In other words, tally an "item census" – to be consistent with the usage of that term.
    """

    def __init__ (
        self,
        name: str = "generic",
        ) -> None:
        """
Constructor for an item census.

    name:
optional name for this measure
        """
        self.name = name
        self.count: dict = defaultdict(int)
        self.df = None


    def increment (
        self,
        item0: Census_Item,
        ) -> None:
        """
Increment the count for this item.

    item0:
an item (domain: node, predicate, label, URL, literal, etc.) to be counted
        """
        self.count[item0] += 1


    def get_tally (
        self
        ) -> typing.Optional[pd.DataFrame]:
        """
Accessor for the item counts.

    returns:
a [`pandas.DataFrame`](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html) with the count distribution, sorted in ascending order
        """
        self.df = pd.DataFrame.from_dict(  # type: ignore
            self.count,
            orient="index",
            columns=["count"],
            ).sort_values("count", ascending=False)
        return self.df


    def get_keyset (
        self
        ) -> set:
        """
Accessor for the set of items (domain) counted.

    returns:
set of keys for the items (domain: nodes, predicates, labels, URLs, literals, etc.) that were counted
        """
        return { key.toPython() for key in self.count.keys() }


class Simplex1 (Simplex0):
    """
Measure a dyad census in an RDF graph, i.e., count the relations (directed edges) which connect two nodes.
    """

    def __init__ (
        self,
        name: str = "generic",
        ) -> None:
        """
Constructor for a dyad census.

    name:
optional name for this measure
        """
        super().__init__(name=name)  # type: ignore
        self.link_map: typing.Optional[dict] = None


    def increment (  # type: ignore # pylint: disable=W0221 # lgtm[py/inheritance/signature-mismatch]
        self,
        item0: Census_Item,
        item1: Census_Item,
        ) -> None:
        """
Increment the count for a dyad represented by the two given items.

    item0:
"source" item (domain: node, label, URL, etc.) to be counted

    item1:
"sink" item (range: node, label, literal, URL, etc.) to be counted
        """
        link = (item0, item1,)
        self.count[link] += 1


    def get_tally_map (
        self
        ) -> Census_Dyad_Tally:
        """
Accessor for the dyads census.

    returns:
a tuple of a [`pandas.DataFrame`](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.html) with the count distribution, sorted in ascending order; and a map of the observed links between "source" and "sink" items
        """
        super().get_tally()
        self.link_map = defaultdict(set)

        for index, _ in self.df.iterrows():  # type: ignore
            item0, item1 = index
            self.link_map[item0].add(item1)

        return self.df, self.link_map  # type: ignore


class Measure:
    """
This class measures an RDF graph.
Its downstream use cases include: graph size estimates; computation costs; constructed shapes.
See <https://derwen.ai/docs/kgl/concepts/#measure>

Core feature areas include:

  * descriptive statistics
  * topological analysis
    """

    def __init__ (
        self,
        *,
        name: str = "generic",
        ) -> None:
        """
Constructor for this graph measure.

    name:
optional name for this measure
        """
        self.name = name
        self.edge_count = 0
        self.node_count = 0
        self.reset()


    def reset (
        self
        ) -> None:
        """
Reset (reinitialize) all of the counts for different kinds of census, which include:

  * total nodes
  * total edges
  * count for each kind of *subject* (`Simplex0`)
  * count for each kind of *predicate* (`Simplex0`)
  * count for each kind of *object* (`Simplex0`)
  * count for each kind of *literal* (`Simplex0`)
  * item census (`Simplex1`)
  * dyad census (`Simplex1`)
        """
        self.edge_count = 0
        self.node_count = 0
        self.s_gen = Simplex0("subject")
        self.p_gen = Simplex0("predicate")
        self.o_gen = Simplex0("object")
        self.l_gen = Simplex0("literal")
        self.n_gen = Simplex1("node")
        self.e_gen = Simplex1("edge")


    def get_node_count (
        self
        ) -> int:
        """
Accessor for the node count.

    returns:
value of `node_count`
        """
        return self.node_count


    def get_edge_count (
        self
        ) -> int:
        """
Accessor for the edge count.

    returns:
value of `edge_count`
        """
        return self.edge_count


    def measure_graph (
        self,
        kg: KnowledgeGraph,
        ) -> None:
        """
Run a full measure of the given RDF graph.

    kg:
`KnowledgeGraph` object representing the RDF graph to be measured
        """
        for s, p, o in kg.rdf_graph():
            self.edge_count += 1
            self.s_gen.increment(s)
            self.p_gen.increment(p)
            self.n_gen.increment(s, p)

            if isinstance(o, rdflib.term.Literal):
                self.l_gen.increment(o)
            else:
                self.o_gen.increment(o)
                self.e_gen.increment(p, o)

        self.node_count = len(set(self.s_gen.count.keys()).union(set(self.o_gen.count.keys())))


    def get_keyset (
        self,
        *,
        incl_pred: bool = True,
        ) -> typing.List[str]:
        """
Accessor for the set of items (domain: nodes, predicates, labels, URLs, literals, etc.) that were measured.
Used for *label encoding* in the transform between an RDF graph and a matrix or tensor representation.

    incl_pred:
flag to include the predicates in the set of keys to be encoded

    returns:
sorted list of keys to be used in the encoding
        """
        keys = self.s_gen.get_keyset().union(self.o_gen.get_keyset())

        if incl_pred:
            keys = keys.union(self.p_gen.get_keyset())

        return sorted(list(keys))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# see license https://github.com/DerwenAI/kglab#license-and-copyright

"""
Provide support for importing RDF data from multiple existing graph
databases, including:

  * neo4j
"""

import json
import urllib.parse

import requests  # pylint: disable=E0401
import rdflib  # type: ignore  # pylint: disable=E0401


def import_from_neo4j (
    username: str,
    password: str,
    dbname: str,  # pylint: disable=W0613
    host: str = "localhost",
    port: str = "7474"
    ) -> rdflib.Graph:
    """
Wrapper for a
[`Cypher`](https://neo4j.com/labs/neosemantics/tutorial/#_using_the_cypher_n10s_rdf_export_procedure)
export request, to provide neo4j integration through the
[`neosemantics`](https://neo4j.com/labs/neosemantics/) library.

Tested with ~10GB of stored triples.

    username:
the user name, as a string

    password:
the password, as a string

    dbname:
the database name, as a string

    host:
optionally, the neo4j server domain name or IP address, as a string – including the protocol scheme; defaults to `"http://localhost"`

    port:
optionally, the neo4j server port; defaults to `"7474"`

    returns:
an [`rdflib.Graph`](https://rdflib.readthedocs.io/en/stable/apidocs/rdflib.html?highlight=graph#graph) object parsed from the exported RDF
    """
    # get all of the entities and properties via `neosemantics`
    # exported as RDF triples
    cypher = "Match (n)-[r]->(m) Return n,r,m;"

    payload = {
        "cypher": cypher,
        "format": "RDF/XML",
        }

    # construct the export URL
    u = urllib.parse.urlparse(host)
    netloc = f"{u.netloc}:{port}"
    path = "/rdf/{dbname}/cypher"
    url = urllib.parse.urlunparse((u.scheme, netloc, path, "", "", "",))

    try:
        response = requests.post(
            url,
            auth = (username, password),
            data = json.dumps(payload),
            timeout = 10,
        )

        # raise an error on unsuccessful status codes
        response.raise_for_status()
    except Exception as e:  # pylint: disable=W0703
        print("neo4j import: ", e)

    g = rdflib.Graph().parse(data=response.text)
    return g


######################################################################
## main entry point

if __name__ == "__main__":
    graph = import_from_neo4j(  # nosec
        username = "user",
        password = "pswd",
        dbname = "rdfdb",
        host = "localhost",
        port = "7474",
        )



"""
Working with `SubgraphMatrix` as vectorized representation.
Additions to functionalities present in `subg.py`.
Integrate `scikit-network` functionalities.

see license https://github.com/DerwenAI/kglab#license-and-copyright
"""

import networkx as nx
from networkx.exception import NetworkXError
from scipy.spatial.distance import pdist, squareform

from .util import Mixin


class NetAnalysisMixin (Mixin):
    """
Provides methods for network analysis tools to work with `KnowledgeGraph`.
    """

    def get_distances (self, adj_mtx):
        """
Compute distances according to an adjacency matrix.

        adj_mtx:
numpy.array: square matrix of distances.
        """
        self.check_attributes()
        return squareform(pdist(adj_mtx, metric="euclidean"))


    def get_shortest_path (self, src, dst):
        """
Return shortest path from sources to destinations.

        src:
int or iterable: indices of source nodes
        dst:
int or iterable: indices of destination nodes

        returns:
list of int: a path of indices
        """
        self.check_attributes()
        return nx.shortest_path(self.nx_graph, source=src, target=dst)


    def describe (self):
        """
Return a summary for subgraph statistics.
NOTE: we may cache these methods calls if we create something like a `GraphFrame` object.
  see kglab#273, same for adjacency and other matrices.

        return:
dict: a dictionary with stats
        """

        def msg_if_raise (f, g, r):
            """Handle error messages by adding a message key in the results"""
            try:
                return f(g)
            except NetworkXError as e:
                r[f"{str(f.__name__)}_msg"] =  str(e)
                return None

        results = {
            "n_nodes": self._get_n_nodes(),
            "n_edges": self._get_n_edges(),
        }

        return { **results, **{
            "center": msg_if_raise(nx.center, self.nx_graph, results),
            "diameter": msg_if_raise(nx.diameter, self.nx_graph, results),
            "eccentricity": msg_if_raise(nx.eccentricity, self.nx_graph, results)
        }}


    def describe_more (self):
        """
Return a summary with more graph statistics.
        """
        # density
        # triangles
        # reciprocity
        raise NotImplementedError()
